package map;

import index.CellBounds;
import index.CellMethod;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;

import jp.sourceforge.ma38su.util.Log;

import map.model.City;
import map.model.Node;
import map.route.RouteNavigation;
import map.store.Store;
import view.DialogFactory;
import view.MapPanel;
import view.StatusBar;
import database.FileDatabase;
import database.CodeDatabase;

/**
 * 市区町村管理クラス
 * @author ma38su
 */
public class CityManager extends Thread {

	/**
	 * セル型の地域検索クラス
	 */
	private final CellMethod cell;

	/**
	 * 市区町村データのMap
	 */
	private final Map<Integer, City> cityMap;
	final CodeDatabase db;
	
	/**
	 * 街区レベル位置参照情報ファクトリー
	 */
	private final IsjReader isjFactory;

	private final String[] name = new String[]{"北海道", "青森県", "岩手県", "宮城県", "秋田県", "山形県", "福島県", "茨城県", "栃木県", "群馬県", "埼玉県", "千葉県", "東京都", "神奈川県", "新潟県", "富山県", "石川県", "福井県", "山梨県", "長野県", "岐阜県", "静岡県", "愛知県", "三重県", "滋賀県", "京都府", "大阪府", "兵庫県", "奈良県", "和歌山県", "鳥取県", "島根県", "岡山県", "広島県", "山口県", "徳島県", "香川県", "愛媛県", "高知県", "福岡県", "佐賀県", "長崎県", "熊本県", "大分県", "宮崎県", "鹿児島県", "沖縄県"};

	/**
	 * 地図パネルクラス
	 */
	final MapPanel panel;

	/**
	 * 都道府県ごとの市区町村の読み込み状態
	 * true  -> 既読
	 * false -> 未読
	 */
	private boolean[] prefecture;

	private Rectangle screen;

	/**
	 * 数値地図2500を読み込むためのクラス
	 */
	private final Sdf2500Reader sdf2500;
	
	/**
	 * 数値地図25000を読み込むためのクラス
	 */
	private final Sdf25kReader sdf25k;

	/**
	 * 店舗情報ファクトリー
	 */
	private final Collection<Store> shops;

	/**
	 * ステータスバー
	 */
	private final StatusBar statusbar;
	
	public CityManager(MapPanel panel, final CellMethod cell, FileDatabase storage, CodeDatabase db, IsjReader isjFactory, Collection<Store> shop, StatusBar statusbar) {
		this.db = db;
		this.cityMap = new HashMap<Integer, City>();
		this.panel = panel;
		this.cell = cell;
		this.shops = shop;
		this.sdf25k = new Sdf25kReader(storage);
		this.sdf2500 = new Sdf2500Reader(storage);
		this.isjFactory = isjFactory;
		this.statusbar = statusbar;
		this.prefecture = new boolean[47];
		this.screen = this.panel.getScreen();
	}

	public void clear() {
		this.cityMap.clear();
	}

	/**
	 * 経路探索のために読み込んだ頂点を破棄します。
	 */
	public synchronized void clearNode() {
		for (City map : this.cityMap.values()) {
			if ((!map.hasData() && map.hasNode()) || (map.hasData() && !this.screen.intersects(map.getArea()))) {
				map.dump();
			}
		}
	}
	
	public boolean containsKey(Object key) {
		return this.cityMap.containsKey(key);
	}
	
	public boolean containsValue(Object value) {
		return this.cityMap.containsValue(value);
	}

	/**
	 * 不要なデータの解放
	 */
	public synchronized void dump() {
		if (this.panel.mode() == 3) {
			for (City map : this.cityMap.values()) {
				if (map.hasData() && !this.screen.intersects(map.getArea())) {
					this.statusbar.startReading("DUMP : "+ this.db.get(map.getCode()));
					map.dump();
					this.statusbar.finishReading();
				}
			}
		} else {
			for (City map : this.cityMap.values()) {
				if (map.hasData()) {
					this.statusbar.startReading("DUMP : "+ this.db.get(map.getCode()));
					map.dump();
					this.statusbar.finishReading();
				}
			}
		}
	}

	/**
	 * 引数の都道府県番号に含まれていないデータの解放
	 * @param codes 都道府県番号
	 */
	public synchronized void dumpPrefecture(Set<Integer> codes) {
		Iterator<Map.Entry<Integer, City>> itr = this.cityMap.entrySet().iterator();
		synchronized (this.cityMap) {
			while (itr.hasNext()) {
				Map.Entry<Integer, City> entry = itr.next();
				int prefCode = entry.getKey() / 1000;
				if (!codes.contains(prefCode)) {
					this.prefecture[prefCode - 1] = false;
					this.statusbar.startReading("DUMP "+ this.name[prefCode - 1]);
					itr.remove();
				}
			}
		}
		this.statusbar.finishReading();
	}

	/**
	 * 市区町村データの取得
	 * @param key 市区町村番号
	 * @return 市区町村データ
	 */
	public City get(Object key) {
		City city = this.cityMap.get(key);
		return city;
	}

	/**
	 * 指定した頂点番号の頂点を取得します。
	 * @param id 頂点番号
	 * @param isDetailRoad 詳細な道路区間データを取得する場合true
	 * @return 頂点
	 * @throws IOException 頂点を取得できなかった場合
	 */
	public Node getNode(long id, boolean isDetailRoad) throws IOException {
		City map = this.cityMap.get((int) (id / 1000000));
		if (map == null) {
			map = this.sdf25k.productNode((int) (id / 1000000), isDetailRoad);
		} else if (!map.hasNode()) {
			this.sdf25k.productNode(map, isDetailRoad);
		}
		Node node = map.getNode((int) (id % 1000000));
		if (node == null) {
			Log.out(this, " : node is not found, and reproduct node.");
			this.sdf25k.productNode(map, isDetailRoad);
			node = map.getNode((int) (id % 1000000));
			if (node == null) {
				throw new IOException("node is not found.");
			}
		}
		return node;
	}
	
	/**
	 * 頂点のCollectionを取得
	 * @param code 市区町村番号
	 * @return 頂点のCollection
	 */
	public Collection<Node> getNodes(int code) {
		return this.cityMap.get(code).getNode();
	}
	/**
	 * とりあえず行政代表点を求めます。
	 * @param code 市区町村番号
	 * @return 市区町村番号に対応した座標
	 */
	public Point getPoint(int code) {
		Point point = null;
		try {
			point = this.sdf25k.getPoint(code);
		} catch (Exception e) {
			point = null;
			Log.err(this, e);
			Log.err(this, "code: "+ Integer.toString(code));
		}
		return point;
	}
	
	public String getPrefecture(int i) {
		return this.name[i];
	}

	/**
	 * 都道府県データ
	 * @param code
	 * @return 既読ならtrue、未読ならfalse
	 */
	public boolean hasPrefecture(int code) {
		return this.prefecture[code - 1];
	}
	
	/**
	 * 国土数値情報の行政界の読み込み市区町村データを生成
	 * @param code 市区町村番号
	 * @throws IOException 入出力エラー
	 */
	void productPrefecture(int code) throws IOException {
		if (!this.prefecture[code - 1]) {
			final List<City> list = new ArrayList<City>();
			String format = City.prefectureFormat(code);
			Map<Integer, List<Polygon>> polygonMap = KsjReader.readSerializeKsj(format);
			for (Map.Entry<Integer, List<Polygon>> entry : polygonMap.entrySet()) {
				Integer key = entry.getKey();
				City city = new City(key, entry.getValue().toArray(new Polygon[] {}));
				list.add(city);
				this.put(key, city);
			}
			this.prefecture[code - 1] = true;
			Thread thread = new Thread() {
				@Override
				public void run() {
					for (City city : list) {
						city.setName(CityManager.this.db.get(city.getCode()));
					}
					CityManager.this.panel.repaint();
				}
			};
			thread.start();
		}
	}

	/**
	 * 地図情報を加える
	 * @param key 市区町村番号
	 * @param value 市区町村データ
	 * 
	 * @return 市区町村データ
	 */
	public City put(Integer key, City value) {
		synchronized (this) {
			return this.cityMap.put(key, value);
		}
	}

	/**
	 * 市町村データの読み込み
	 * @param code 読み込む市区町村番号
	 * @param rect 
	 * @return データを読み込めばtrue
	 * @throws IOException 入出力エラー
	 */
	private boolean readCity(int code, Rectangle rect) {
		Log.out(this, "read City started.");
		// セルメソッドから地図の読み込み
		Log.out(this, "read City: " + code);
		City city = this.get(code);
		Log.out(this, "city: " + city);
		boolean ret = false;
		String name = city.getName();
		if (name == null) {
			name = this.db.get(code);
			city.setName(name);
		}
		String content = "READ "+ name;
		if (!rect.intersects(this.screen)) {
			return false;
		}
		if (city == null) {
			/* 数値地図25000を読み込む */
			this.statusbar.startReading(content, StatusBar.SDF25000);
			city = this.sdf25k.product(code);
			this.put(code, city);
			Log.out(this, "put " + code + ", " + city);
			this.panel.repaint();
			ret = true;
		} else if (!city.hasData()) {
			/* 数値地図25000を読み込む */
			Log.out(this, "read SDF25000: "+ code);
			this.statusbar.startReading(content, StatusBar.SDF25000);
			this.sdf25k.product(city);
			this.panel.repaint();
			ret = true;
		}
		if (!this.screen.intersects(city.getArea())) {
			return false;
		}
		if (!city.hasStore()) {
			/* 店舗情報を読み込む */
			Log.out(this, "read ISJ: "+ code);
			this.statusbar.startReading(content, StatusBar.ISJ);
			Map<String, Point> locationMap = this.isjFactory.productStreaming(code);
			// TODO 富士山頂付近でnullが返ることがある。
			if (locationMap != null) {
				for (Store shop : this.shops) {
					try {
						List<Point> location = shop.getLocation(city, locationMap);
						city.addLabels(shop.getName(), location);
					} catch (Throwable e) {
						DialogFactory.errorDialog(this.panel, e);
					}
				}
			}
			city.setStore();
			this.panel.repaint();
			ret = true;
		}
		if (!this.screen.intersects(city.getArea())) {
			return false;
		}
		if (!city.isSdk2500() && this.db.hasSdf2500(code)) {
			/* 数値地図2500を読み込む */
			Log.out(this, "read SDF2500: "+ code);
			this.statusbar.startReading(content, StatusBar.SDF2500);
			this.sdf2500.readMizu(city);
			this.sdf2500.readGyousei(city);
			this.panel.repaint();
			ret = true;
		}
		this.statusbar.finishReading();
		return ret;
	}

	@Override
	public void run() {
		while (true) {
			Log.out(this, "running...");
			Set<Integer> prefSet = new HashSet<Integer>();
			try {
				if (this.panel.mode() > 1) {
					Rectangle rect;
					do {
						rect = new Rectangle(this.screen);
						SortedMap<CellBounds, Integer> codes = this.cell.search2(rect);
						for (int val : codes.values()) {
							prefSet.add(val / 1000);
						}
						Log.out(this, "codes = " + codes);
						if (!this.panel.isOperation()) {
							this.dumpPrefecture(prefSet);
						}
						for (int val : prefSet) {
							this.productPrefecture(val);
						}
						if (this.panel.mode() == 3) {
							for (Integer code : codes.values()) {
								if (!rect.equals(this.screen)) {
									break;
								}
								Log.out(this, "read KSJ border:"+ City.cityFormat(code));
								this.readCity(code, rect);
							}
						}
					} while (!this.screen.equals(rect));
				}
				synchronized (this.cell) {
					Log.out(this, "wait");
					this.cell.wait();
				}
			} catch (Exception e) {
				Log.err(this, e);
			}
		}
	}

	public void searchedRecover() {
		this.panel.repaint();
		this.clearNode();
	}

	public void set(RouteNavigation navi) {
		this.statusbar.set(navi);
	}
	public Collection<City> values() {
		return this.cityMap.values();
	}

	public void wakeup() {
		synchronized (this.cell) {
			this.cell.notifyAll();
		}
	}
}
